******************************************************************************
* Copyright 1989,1990 David Pomerantz  
* 
* This is the source for the palette translation program described in the
* article "A Few Good Colors" (Computer Language, August 1990).  This code
* was written for Digital Vision Inc.'s ComputerEyes/Pro video digitizer.
* Dave Pratt and Vivian Russell have graciously given permission for 
* this code to be made available to Computer Language readers.
*
* Please include this notice if you copy the program.  Thanks.
*
* If you have questions, I can be reached at:
* 
* Micro Alliance, Inc.		Phone: (617) 245-8879
* 107 Greenwood Street		AppleLink: D6543
* Wakefield, MA  01880		Certified Apple Developer
*****************************************************************************
* FILE: RGBPALET.A
* DESC: Routines to calculate a 'best' palette of 240 colors from 32000 colors
*		collected during image acquisition.
* NOTE:	This is intended for compilation on a Macintosh system under MPW.  It was 
*		not written for portability, but it could be ported to other 68000 systems
*		with a little work.  As far as I know, it is 32-bit clean.
*
* CALLING FROM C:
*
* I call these assembler routines with code that looks roughly like this:
*
* typedef struct {
* 	long	first, last;		/* position of rect within the colorList	*/
*	long	histSum;			/* sum of hist entries in that range	*/
*	byte	red,green,blue;		/* the representative color, palette entry	*/
*	byte	redMin, greenMin, blueMin;
*	byte	redMax, greenMax, blueMax;
*	byte	widestColor;
*	byte	okRange;
*	byte	paletteIndex;
* } COLORRECT;
*
*   COLORRECT   *colorRect = nil;
*	short		*colorList = nil;
*	long		 listLen;
*	short		 entries;
*
*   /* collect a histogram in an array of 32K words (64K bytes) */
*	hist = yourRoutine();		/*  I do this during image capture */
*
*	/* get a list of those colors appearing in the image */
*	if ((colorList = myNewPtr(MAXCOLORS * sizeof(short))) == nil)
*		/* handle error */;
o
*	listLen = fillColorList(hist, colorList); 
*	SetPtrSize((Ptr) colorList, listLen * sizeof(short));
*
*	if ((colorRect = myNewPtr(240 * sizeof(COLORRECT))) == nil)
*		/* handle error */;
*		
*   /* Build the list of color rectangles */
*	entries = buildColorRect(hist, colorRect, colorList, listLen);
*
* 	/* Calculate the weighted average RGB color for each one */
*	if (calcCRcolor() != OK)
*		/* handle error */;
*
*	/* Dispose of now-unneeded memory */
*	if (colorList != nil)
*		myDisposPtr((Ptr *) &colorList);
*
*	/* Calculate the palette lookup-table from the rectangles in colorRect */
*	if ((plut=myNewPtr(MAXCOLORS)) == nil)
*		/* handle error */;
*	bldPlut(plut); 
******************************************************************************
wholeSystem EQU 1
			PRINT	OFF
			CASE	ON
			OPT		NOCLR
			STRING	C
				
			INCLUDE 'Traps.a'
			INCLUDE 'SysEqu.a'
   			INCLUDE 'QuickEqu.a'
			INCLUDE 'SysErr.a'
			
			PRINT	ON,NOMDIR			
			
*****************************************************************************
* NAME: CALL MACRO
* DESC: Call a routine, using the C-language parameter passing conventions
*****************************************************************************
				MACRO
				CALL	&rtn,&argList
				LCLA	&nArg,&pos,&size
				LCLC	&type,&expr
&nArg			SETA	&NBR(&SYSLIST)-1
	
				IMPORT 	(&rtn):CODE
			
		
&size			SETA	0
				WHILE	(&nArg >= 1) DO
&pos			SETA	&SCANEQ(':',&SYSLIST[&nArg+1],1)
&type			SETC	&SUBSTR(&SYSLIST[&nArg+1],&pos+2,1)
&expr			SETC	&SUBSTR(&SYSLIST[&nArg+1],1,&pos)
.				IF		(&type = 'b') goto .byte
				IF		(&type = 'l') goto .long
				IF		(&type = 'a') goto .addr
				IF		(&type = 's') goto .string
&size			SETA	&size+2
				move.w	&expr,-(a7)
				goto	.loop
.addr							
&size			SETA	&size+4
				pea		&expr
				goto	.loop
.byte							
&size			SETA	&size+1
				move.b	&expr,-(a7)
				goto	.loop
.long							
&size			SETA	&size+4
				move.l	&expr,-(a7)
				goto	.loop
.string
&size			SETA	&size+4
				pea		&expr
.loop
&nArg			SETA	&nArg-1
				ENDWHILE
				
				jsr		&rtn
				IF		(&size = 0) goto .exit
				adda	#&size,a7
.exit
				ENDM

MAXCOLORS	EQU		32768	; Max colors in histogram
OK			EQU		0		; success return code
ERR			EQU		-1		; failure return code

;************* COLOR RECTANGLE STRUCTURE DEFINITION ************

colorRect	RECORD	0
first		ds.l	1
last		ds.l	1
histSum		ds.l	1
red			ds.b	1
green		ds.b	1
blue		ds.b	1
redMin		ds.b	1
greenMin	ds.b	1
blueMin		ds.b	1
redMax		ds.b	1
greenMax	ds.b	1
blueMax		ds.b	1
widestColor	ds.b	1		; 0, 1, or 2 for (R, G, or B)
okRange		ds.b	1
palIndex	ds.b	1
sizeCR		equ		*
			ENDR
			
MAXCR		EQU		240		; Max color rect entries
			
locals		RECORD
listAddr	ds.l	1
rectAddr	ds.l	1
histAddr	ds.l	1
numCRused	ds.w	1
numCRavail	ds.w	1
redSelect	ds.b	1				; These are 1-bit switches that make sure that
grnSelect	ds.b	1				;   we alternate evenly between red/blue/green
bluSelect	ds.b	1				;   color ranges.
crOffset	ds.w	MAXCR			; This is a list of offsets into the colorRect
									;  array.  Always kept sorted by histSum.
			ENDR

medianMask	RECORD
			dc.w	$7C00			; 0RRR RR00 0000 0000   Table for isolating
			dc.w	$03E0			; 0000 00GG GGG0 0000     one component from
			dc.w	$001F			; 0000 0000 000B BBBB     a 15-bit color
			ENDR

medianRotate RECORD					; This table rotates a 15-bit color to put a given
			dc.b	10				;   component in the low 5-bits
			dc.b	5
			dc.b	0
			ENDR
			
sortRotate	RECORD
			dc.w	0				; 0RRR RRGG GGGB BBBB   Table for rotating 15-bit 
			dc.w	6				; GGGG GBBB BB0R RRRR     color to make the given
			dc.w	11				; BBBB B0RR RRRG GGGG     component highest in 
			ENDR					;                         sorting priority
			
Rect		RECORD	0
top			ds.w	1
left		ds.w	1
bottom		ds.w	1
right		ds.w	1
			ENDR

			SEG		'Palette'

*****************************************************************************
* NAME: fillColorList
* DESC: Loop through the histogram and transfer the indices (which are the
*		5-bit RGB numbers) to the colorList, for all non-zero entries.
* CDEF:	long fillColorList( short *hist, short *colorList )
* RETN:	the number of colorList entries
* REGS: 
*****************************************************************************
fillColorList	PROC	EXPORT
			
frame		RECORD	{A6Link},DECR
colorList	ds.l	1			; pointer to allocated space (at least 32768 shorts)
hist		ds.l	1			; pointer to 32768 hist entries
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			WITH	frame
			link	a6,#size
			movem.l	d1-d7/a0-a4,-(sp)
			
			move.l	hist(a6),a0
			move.l	colorList(a6),a1
			clr.l	d0						; d0 is # of non-zero entries
			clr.l	d1						; d1 is hist index of each entry
			
fillLoop:
			cmp.w	#0,(a0)+				; if hist entry is nonzero
			beq.s	nullEntry				
			move.w	d1,(a1)+				;   its index is our RGB number
			addq.l	#1,d0					;   save in colorList
nullEntry:
			addq.l	#1,d1					; ++index
			cmp.l	#MAXCOLORS,d1			
			blt.s	fillLoop				;      

			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC

*****************************************************************************
* NAME: buildColorRect
* DESC: 
* CDEF:	
* RETN:	number of cr entries used
* REGS: 
*****************************************************************************
buildColorRect		PROC	EXPORT
			IMPORT	(doBuildRect):CODE
			
frame		RECORD	{A6Link},DECR
listLen		ds.l	1
listAddr	ds.l	1
rectAddr	ds.l	1
histAddr	ds.l	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			WITH	frame
			link	a6,#size
			movem.l	d1-d7/a0-a4,-(sp)
			
			move.l	listAddr(a6),locals.listAddr
			move.l	rectAddr(a6),locals.rectAddr
			move.l	histAddr(a6),locals.histAddr
			
			move.l	listAddr(a6),a2
			move.l	rectAddr(a6),a3
			move.l	histAddr(a6),a4
						
			WITH	locals
			move.w	#1,numCRused			; 1 entry used, MAXCR-1 left
			move.w	#(MAXCR-1),numCRavail	; The first offset in the crOffset
			move.w	#0,crOffset				;   points to the first CR entry
			
			move.b	#0,redSelect			; Initialize preferences for blue,green,
			move.b	#0,grnSelect			;   or red widest colors.  We try to vary
			move.b	#0,bluSelect			;   which ones we pick when the ranges
											;   are equal.
			move.l	#0,colorRect.first(a3)	; We'll initialize the first CR entry
			move.l	listLen(a6),d0			;   as the entire histogram space
			subq.l	#1,d0
			move.l	d0,colorRect.last(a3)
			
			move.b	#$00,colorRect.redMin(a3)
			move.b	#$00,colorRect.greenMin(a3)
			move.b	#$00,colorRect.blueMin(a3)
			move.b	#$1F,colorRect.redMax(a3)
			move.b	#$1F,colorRect.greenMax(a3)
			move.b	#$1F,colorRect.blueMax(a3)
			
			call 	getListRange,#0:w,#4:w,#0:w		; fill in ranges for this entry

loop			
			move.l	rectAddr,a3
			move.w	crOffset,d3		
			cmp.b	#1,colorRect.okRange(a3,d3.w)
			bne.s	endLoop
			cmp.w	#0,numCRavail			; If we can do another median cut
			ble.s	endLoop					;   (i.e., more entries available)
			call	medianCut,d3:w
			bra.s	loop
endLoop
			clr.l	d0
			move.w	locals.numCRused,d0
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC
			
*****************************************************************************
* NAME: medianCut
* DESC: This routine takes a colorRect entry as input, calculates the color 
*		ranges within that rectangle, and sorts the color entries on the 
*		largest range.
*		It then splits the rectangle at the weighted median, and calls itself
*		to recursively split the daughter rectangles.  This splitting goes
*		on until we've created 240 rectangles, or until all rectangles have
*		fewer than three color entries.
* RETN:	
* REGS: 
*		d3		offset into colorRect for current entry
*		a2		INPUT:	colorList array
*		a3		INPUT:	colorRect array
*		a4		INPUT:	histogram array
*****************************************************************************
medianCut	PROC	ENTRY
			
frame		RECORD	{A6Link},DECR
offset		ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size	
			movem.l	d1-d7/a0-a4,-(sp)
			
			move.w	frame.offset(a6),d3
			
			WITH	colorRect
			
			clr.w	d1						; The widestColor is a number from 0 to 2
			move.b	widestColor(a3,d3.w),d1	;   representing the color with the greatest
			lsl.w	#1,d1					;   variation.  We'll sort on that color by
			lea		medianMask,a0			;   rotating each color value either 0, 6, or
			move.w	0(a0,d1.w),d2			;   11 bits.  The sort mask allows us to 
											;   isolate that color when we look for the
			lea		sortRotate,a0			;   median in a range of colors.
			move.w	0(a0,d1.w),d1
						
			call	quickSort,d1:w,first(a3,d3.w):l,last(a3,d3.w):l
			
			call	findMedian,d2:w,d3:w	; Get median value into d0
			call	splitCR,d3:w,d0:l		; Split at median value into two colorRect's
											;   put new CR offset into d0
			move.w	d0,d2
			call	removeCR,d3:w
			call	insertCR,d3:w
			call	insertCR,d2:w
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC
*****************************************************************************
* NAME: getListRange
* DESC: Fill in these values in the colorRect entry:
*		   redMax, redMin, greenMax, greenMin, blueMax, blueMin
*		   histSum, widestColor
*		All these values must be present for the colorRect entry to be turned
*		into a palette entry.  The max/min values determine the rectangle within
*		the color cube.  The histSum tells how many pixels are in this rectangle.
*		The widestColor tells which color has the longest edge of the rectangle.
* RETN:	
* REGS: max and min values for each color are kept in register words
*		The max is kept in the low 8 bits, the min is kept in the high 8 bits
*		d2		red min/max
*		d3		green min/max
*		d4		blue min/max
*		d5		histogram sum
*		d6		offset into colorList
*		d7		offset of end of range
*
*		a2		INPUT: colorList
*		a3		INPUT: top of colorRect array (we adjust to point to current entry)
*		a4		INPUT: histogram
*****************************************************************************
			MACRO	
			cmpMinMax &data,&minmax
			LCLC	&maxset,&minset
&maxset		SETC	&CONCAT('maxset',&SYSNDX)
&minset		SETC	&CONCAT('minset',&SYSNDX)
			andi.b	#$1F,&data
			cmp.b	&data,&minmax
			bge.s	&maxset
			move.b	&data,&minmax
&maxset
			rol.w	#8,&minmax
			cmp.b	&data,&minmax
			ble.s	&minset
			move.b	&data,&minmax
&minset
			ror.w	#8,&minmax
			ENDM
			
getListRange	PROC	ENTRY
			
frame		RECORD	{A6Link},DECR
face		ds.w	1
color		ds.w	1
offset		ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
size	equ		*
			ENDR

			link	a6,#frame.size
			movem.l	d1-d7/a3,-(sp)
			
			adda.w	frame.offset(a6),a3
			
			WITH	colorRect
			clr.l	d5
			clr.l	d0
			move.l	first(a3),d6
			move.l	last(a3),d7

			move.b	#1,okRange(a3) 		; ? enough entries to split? assume yes
			move.l	d7,d2				; if (last - first) >= 1
			sub.l	d6,d2				;   then we could reasonably split this
			cmp.l	#1,d2				;   entry
			bge.s	enough					
			move.b	#0,okRange(a3)
enough
			lsl.l	#1,d6				; Now, multiply first and last by two
			lsl.l	#1,d7				;  so they represent word offsets
			
			move.w	#$1F00,d2
			move.w	#$1F00,d3
			move.w	#$1F00,d4
			
loop:
			move.w	0(a2,d6.l),d1				; get color value
			lsl.l	#1,d1						;  times two to get index into histogram
			move.w	0(a4,d1.l),d0				; get hist value for current color value
			add.l	d0,d5						; convert to long and add to total sum
			lsr.l	#1,d1
			
			move.b	d1,d0						; Get blue component
			cmpMinMax	d0,d4					;   set min and max
			
			lsr.w	#5,d1						; Get green component
			move.b	d1,d0						;   set min and max
			cmpMinMax	d0,d3
			
			lsr.w	#5,d1						; Get red component
			move.b	d1,d0						;   set min and max
			cmpMinMax	d0,d2
			
			addq.l	#2,d6						; loop through all colors in the
			cmp.l	d7,d6						;   given region (between first and last)
			ble.s	loop
			
			move.l	d5,histSum(a3)
			
			clr.w	d0							; get min/max for red
			move.b	d2,d0						;   d0 (red range) = max - min
			ror.w	#8,d2
			andi.w	#$00FF,d2
			sub.b	d2,d0
			
			clr.w	d1							; get min/max for green
			move.b	d3,d1						;   d1 (green range) = max - min
			ror.w	#8,d3
			andi.w	#$00FF,d3
			sub.b	d3,d1
			
			clr.w	d2							; get min/max for blue
			move.b	d4,d2						;   d2 (blue range) = max - min
			ror.w	#8,d4
			andi.w	#$00FF,d4
			sub.b	d4,d2
			
; Now: d0, d1, and d2 have the color ranges for red, green and blue.
; At this point we want to set widestColor to either 0, 1, or 2, depending
; on which of the three ranges is widest.  Since there are only 32 shades
; of each color component, it's common for two ranges to be the same.  
; So, we want to alternate which component we say is widest, when two or
; more components have the same range.

			cmp.b	d0,d1				; ? green bigger than red?
			bhi.s	greenWider			; yup, test blue
			beq.s	greenEqualRed		; equal, choose between 'em.		
			cmp.b	d0,d2				; ? red bigger than blue?
			bhi.s	chooseBlue			; yup, choose blue
			beq.s	blueEqualRed		; equal, choose between 'em
chooseRed
			move.b	#0,widestColor(a3)	; choose red
			bra.s	exit
greenWider				
			cmp.b	d1,d2				; ? blue bigger than green ?
			bhi.s	chooseBlue			; yup, choose it
			beq.s	blueEqualGreen		; equal, pick between 'em
chooseGreen
			move.b	#1,widestColor(a3)	; choose green
			bra.s	exit
chooseBlue
			move.b	#2,widestColor(a3)
			bra.s	exit
			
greenEqualRed				
			cmp.b	d0,d2				; blue bigger than green/red?
			bhi.s	chooseBlue			; yup, choose blue			
			beq.s	blueEqualRed		; equal, pick between r,g,b
			eori.b	#1,locals.redSelect	; was red picked last time?
			beq.s	chooseGreen			; yup, choose green this time
			bra.s	chooseRed			; nope, choose red this time
blueEqualRed
			eori.b 	#1,locals.bluSelect	; was red picked last time?
			beq.s	chooseBlue			; yup, choose blue this time
			bra.s	chooseRed			; nope, choose red this time
blueEqualGreen
			eori.b 	#1,locals.grnSelect	; was green chosen over blue last time?
			beq.s	chooseBlue			; yup, choose blue this time
			bra.s	chooseGreen			; nope, choose green this time
			
exit
			movem.l	(sp)+,d1-d7/a3
			unlk	a6
			rts
			ENDPROC

*****************************************************************************
* NAME: findMedian
* DESC: Normally, a median point is the point where the sum of the histogram
*		entries is half the total sum of the entries.  We need to do a little
*		better, and use only those points where the color has changed.
*
*		So, we add the histogram entries as we proceed from first to last.
*		When the color changes (we're only interested in the color that we
*		sorted on), we look at the difference between 1/2 the histogram total
*		and the present total.  We work to minimize that difference.
* CDEF:	
* RETN:	median point is returned in d0.l
* REGS:  
*       d0		index of current hist entry (returned as median pt.)
*		d1		index of last hist entry
*		d2		half of sum of hist entries in this range
*		d3		current sum of entries to this point
*
*		a2		INPUT: colorList
*		a3		INPUT: top of colorRect array (we adjust to point to current entry)
*		a4		INPUT: histogram
*                   
*****************************************************************************
findMedian	PROC	ENTRY

frame		RECORD	{A6Link},DECR
offset		ds.w	1
colorMask	ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
lastIndex	ds.l	1
lastDiff	ds.l	1
lastColor	ds.w	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a3,-(sp)
			
			adda.w	frame.offset(a6),a3
			WITH	colorRect
			
			move.l	first(a3),d0				; d0 -> first hist entry
			move.l	last(a3),d1					; d1 -> last hist entry
			lsl.l	#1,d0						; hist entries are two
			lsl.l	#1,d1						;   bytes (word values)
			move.l	histSum(a3),d2				; get sum over the range
			lsr.l	#1,d2						; divide by two for median sum
			
			WITH	frame
			move.w	#$FFFF,lastColor(a6)		; this value differs from all colors
			move.l	d0,lastIndex(a6)			; in a pinch, this is the median
			move.l	#$7FFFFFFF,lastDiff(a6)		; a very big difference	
			
			clr.l	d3							; clear accumulated sum
			clr.l	d5							; word -> long conversion
			
			WITH	frame
loop:
			clr.l	d4
			move.w	0(a2,d0.l),d4				; get color value
			lsl.l	#1,d4						;  times two to get byte offset into histogram
			move.w	0(a4,d4.l),d5				; get hist value for current color value
			add.l	d5,d3						; convert to long and add to running total
			
			lsr.l	#1,d4						; back to normal color value
			and.w	colorMask(a6),d4			; isolate the color on which we're sorted
			
			cmp.w	lastColor(a6),d4			; ? Has it changed ?
			beq.s	noMedian					;   no, we don't have a median
			cmp.w	#$FFFF,lastColor(a6)		; ? first item ?
			beq.s	noMedian					;   can't use the first item
			move.l	d3,d6						;        get running total
			sub.l	d2,d6						;        subtract range total
			bpl.s	isPositive					;		 get
			neg.l	d6							;          abs
isPositive										;            value
			cmp.l	lastDiff(a6),d6				;        ? smaller than last diff ?
			bgt.s	noMedian					;          no, we want smallest diff
			move.l	d6,lastDiff(a6)				;          yes, we'll take it
			move.l	d0,lastIndex(a6)			;               and its word offset
noMedian
			move.w	d4,lastColor(a6)			; save the color
			addq.l	#2,d0						; next entry
			cmp.l	d1,d0						; ? at end ?
			ble.s	loop						;   no, keep going

			move.l	lastIndex(a6),d0			; get the byte offset of the median pt.
			lsr.l	#1,d0						; make it a word offset
			cmp.l	colorRect.first(a3),d0		; ? is it same as the first entry ?
			ble.s	exit						;   yes, leave it be
			subq.l	#1,d0						;   less one, for the end of the low range
exit		
			movem.l	(sp)+,d1-d7/a3				; d0 = median point
			unlk	a6
			rts
			ENDPROC

*****************************************************************************
* NAME: splitCR
* DESC: Split the colorRect entry at the median point
* CDEF:	
* RETN:	
* REGS: 
*		a2		INPUT: colorList
*		a3		INPUT: top of colorRect array 
*		a4		INPUT: histogram
*****************************************************************************
splitCR		PROC	EXPORT
			
frame		RECORD	{A6Link},DECR
median		ds.l	1
offset		ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0,-(sp)
			
			WITH	colorRect
			move.w	locals.numCRused,d4		; d4 is offset to the new
			mulu	#sizeCR,d4				;  entry
			move.w	frame.offset(a6),d3		; d3 is offset to old entry
			
			clr.w	d2						; d2 has widestColor index
			move.b	widestColor(a3,d3.w),d2 ;   we are cutting this entry across
											;     the color where max-min is greatest
											
****
* Set up the new colorRect entry: first, last
* This entry will be the low-side of the cutpoint.
****
			move.l	first(a3,d3.w),first(a3,d4.w)
			move.l	frame.median(a6),last(a3,d4.w)
			move.b	redMin(a3,d3.w),redMin(a3,d4.w)
			move.b	blueMin(a3,d3.w),blueMin(a3,d4.w)
			move.b	greenMin(a3,d3.w),greenMin(a3,d4.w)
			move.b	redMax(a3,d3.w),redMax(a3,d4.w)
			move.b	blueMax(a3,d3.w),blueMax(a3,d4.w)
			move.b	greenMax(a3,d3.w),greenMax(a3,d4.w)
			
****
* Only one max/min pair changes.  Change the min for the high region and
* change the max for the low region, at the widestColor.
****
			move.l	locals.listAddr,a2		; list of colors
			move.l	frame.median(a6),d0		; index of median color
			asl.l	#1,d0					;   times two for byte offset
			move.w	0(a2,d0.l),d0			; d0 = 15-bit median color
			lea		medianRotate,a0			; table of rotation values
			clr.l	d1						; get number of bits
			move.b	0(a0,d2.w),d1			;   to isolate widestColor
			tst.w	d1
			beq.s	skipRotate
			ror.w	d1,d0					; color in low 5 bits
skipRotate
			andi.b	#$1F,d0
			
			lea		redMax(a3,d4.w),a0		; a0->max's for low region
			move.b	d0,0(a0,d2.w)			; change max color for low region
			
			add.b	#1,d0					; next color for min of high region
			lea		redMin(a3,d3.w),a0		; a0->min's for high region
			move.b	d0,0(a0,d2.w)			; change min color for high region
			
			move.l	frame.median(a6),d0		; This entry starts one past the
			addq.l	#1,d0					;   end of the low entry
			move.l	d0,first(a3,d3.w)		; Last remains the same
			
			call 	getListRange,d3:w,d2:w,#0:w		; Calculate range of high entry
			call	getListRange,d4:w,d2:w,#1:w		;   and low entry
			
			
			move.w	d4,d0					; Return offset of the new entry
			
			movem.l	(sp)+,d1-d7/a0
			unlk	a6
			rts
			ENDPROC

*****************************************************************************
* NAME: removeCR
* DESC: Remove a colorRect entry from its position in the offset table
* RETN:	
* REGS: 
*		a3		INPUT:	addr of colorRect table
*****************************************************************************
removeCR	PROC	EXPORT
			
frame	RECORD	{A6Link},DECR
offset		ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0-a4,-(sp)
			
			move.w	frame.offset(a6),d2				; d1 is offset to the colorRect entry
			lea		locals.crOffset,a0				; a0 -> list of offsets of CR's
			move.w	#0,d3
			move.w	locals.numCRused,d1
			subq.w	#1,d1
			asl.w	#1,d1
loop		
			cmp.w	d1,d3
			bgt.s	exit
			cmp.w	0(a0,d3.w),d2
			beq.s	closeUpLoop
			addq.w	#2,d3
			bra.s	loop
			
closeUpLoop	
			cmp.w	d1,d3
			bge.s	endLoop
			move.w	2(a0,d3.w),0(a0,d3.w)			; close up the space
			addq.w	#2,d3
			cmp.w	d1,d3
			ble.s	closeUpLoop
endLoop			
			subq.w	#1,locals.numCRused		; One less colorRect used
			addq.w	#1,locals.numCRavail	;  means one more available

exit
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC			
*****************************************************************************
* NAME: insertCR
* DESC: Insert a colorRect entry in a sorted position in the crOffset list
* RETN:	
* REGS: 
*		a3		INPUT:	addr of colorRect table
*****************************************************************************
insertCR	PROC	EXPORT
			
frame	RECORD	{A6Link},DECR
offset		ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0-a4,-(sp)
			
			move.w	frame.offset(a6),d2				; d2 is offset to the colorRect entry
			
			lea		locals.crOffset,a0				; a0 -> list of offsets of CR's
			move.w	locals.numCRused,d1				; d1 is offset of end of crOffset	
			subq.w	#1,d1
			asl.w	#1,d1
			
			call	findNewPos,d2:w					; get offset into crOffset
													;   of our new entry
			move.w	d0,d3
			
loop		
			cmp.w	d3,d1							; ? gotten to our pos in crOffset ?
			blt.s	endLoop							;   yes, we're done
			move.w	0(a0,d1.w),2(a0,d1.w)			; open up some space
			subq.w	#2,d1
			bra.s	loop
endLoop
			move.w	d2,0(a0,d3.w)					; insert our entry
			addq.w	#1,locals.numCRused				; One more colorRect used
			subq.w	#1,locals.numCRavail			;  means one less available
			
exit
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC		
*****************************************************************************
* NAME: findNewPos
* DESC: Find a new position for a colorRect entry in the crPtr table.  This
* RETN:	d0 = offset into the crOffset array (note: NOT offset into colorRect)
* REGS: 
*		a3		INPUT:	addr of colorRect table
*****************************************************************************
findNewPos	PROC	EXPORT
			
frame	RECORD	{A6Link},DECR
offset		ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0-a4,-(sp)
			
			move.l	locals.rectAddr,a3	
			lea		locals.crOffset,a0
			move.w	frame.offset(a6),d0				; d0 is offset to the colorRect entry
			move.l	colorRect.histSum(a3,d0.w),d1	; d1 = its histSum
			cmp.b	#0,colorRect.okRange(a3,d0.w)	; if sufficient range to split, 
			bne.s	setupLoop						;   find out where to put it
			move.w	locals.numCRused,d2
			asl.w	#1,d2
			bra.s	exit
setupLoop			
			move.w	#0,d2
			move.w	locals.numCRused,d3
			subq.w	#1,d3
			asl.w	#1,d3
						
loop
			cmp.w	d3,d2
			bgt.s	exit
			move.w	0(a0,d2.w),d4
			cmp.l	colorRect.histSum(a3,d4.w),d1
			bgt.s	exit
			cmp.b	#0,colorRect.okRange(a3,d4.w)
			beq.s	exit
			
continue
			addq.w	#2,d2
			bra.s	loop
			
exit
			move.w	d2,d0
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC	
			
*****************************************************************************
* NAME: quickSort
* DESC: 
* CDEF:	
* RETN:	
* REGS: 
*****************************************************************************
quickSort	PROC	ENTRY
			
frame	RECORD	{A6Link},DECR
last		ds.l	1
first		ds.l	1
rotate		ds.w	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			WITH	frame
			link	a6,#size
			movem.l	d1-d3,-(sp)
			
			move.w	rotate(a6),d1
			move.l	first(a6),d2
			move.l	last(a6),d3
			lsl.l	#1,d2
			lsl.l	#1,d3
			clr.l	d0		; must be cleared for longword compares
			clr.l	d7
			clr.l	d4
			call	myqsort

exit			
			movem.l	(sp)+,d1-d3
			unlk	a6
			rts
			ENDPROC

*****************************************************************************
* NAME: qsort
* DESC: Assembler implementation of the Quicker Sort algorithm.  See the
*		article in Computer Language, March 1988, page 56 by Carl W. Reynolds.
*		For reference, the C implementation is shown as comments on the right
*		side.
* ARGS: a2		points to the array (a word array)
*		d1		comparison mask (see note)
*		d2		index of first item
*		d3		index of last item
* REGS: I've done the sneaky trick of equating the regs to the variable names.
*		I found it easier to read that way.  All accesses to data are to 
*		registers, except when the 'array' register is referenced.
* NOTE:	Parameters are not passed on the stack because this is a recursive 
*		routine and I wanted to make it FAST!  Less pushing and popping, and
*		less stack space. 16 bytes per call.
*		
*		Comparisons are between word values, on the bits masked by register d1. 
*		For our purposes here, we're comparing words that have 15 bit RGB values:
*		0RRRRRGGGGGBBBBB.  The mask has either $7C00, $03E0, or $001F.
*		
*		Of course, passing a mask of $FFFF compares on a full 16-bit word.
*
* -----------------------------------------------------------------------------
* 
*			HERE'S THE FULL TEXT OF THE C CODE
*
* void quickSort(short *array, short first, short last) 
* {
*	short 	temp;
*	short	i,j;
*	
*	i = first;
*	if (i < last) {
*		while ((array[i] < array[i+1]) && (i < last))
*			++i;
*	}
*	if (i == last)
*		return;
*	j = last;
*	temp = array[i];
*	
*	do {
*		while ((i != j) && (temp < array[j]))
*			--j;
*		if (i != j) {
*			array[i] = array[j];
*			++i;
*			while ( (i != j) && (array[i] < temp) )
*				++i;
*			if (i != j) {
*				array[j] = array[i];
*				--j;
*			}
*		}
*	} while (i != j);
n
*	
*	array[i] = temp;
*	if (first < j) 
*		quickSort(array, first, j-1);
*	if (i < last)
*		quickSort(array, i+1, last);
* }
*****************************************************************************
myqsort		PROC	ENTRY
			
rotate		equ		d1
first		equ		d2
last		equ		d3
temp		equ		d4
i_indx		equ		d5
j_indx		equ		d6
temp_indx	equ		d7
array		equ		a2
			
			movem.w	d2-d7,-(sp)
			
			clr.l	i_indx
			clr.l	j_indx
			move.w	first,i_indx			; i = first;
			cmp.w	last,i_indx				; if (i < last) {
			bge.s	endWhile1
while1:
			move.w	2(array,i_indx.l),d0
			move.w	0(array,i_indx.l),d7
			tst.b	rotate
			beq.s	skip1
			rol.w	rotate,d0
			rol.w	rotate,d7
skip1
			cmp.w	d0,d7
			bhs.s	endWhile1				; 	 while (A[i] < A[i+1]) 
			cmp.w	last,i_indx				;          && (i < last) {
			bge.s	endWhile1
			addq.w	#2,i_indx				;		++i;
			bra.s	while1					;    }
endWhile1:									; }
			cmp.w	i_indx,last				; if (i == last)
			beq.s	exit					;	return;
			
			move.w	last,j_indx				; j = last;
			move.w	0(array,i_indx.l),temp    ; temp = A[i]
			move.w	temp,temp_indx		
			tst.b	rotate
			beq.s	skip2
			rol.w	rotate,temp
skip2
			
doWhile:									; do {
while2:		cmp		i_indx,j_indx			;     while ( (i != j)
			beq.s	endWhile2
			move.w	0(array,j_indx.l),d0
			tst.b	rotate
			beq.s	skip3
			rol.w	rotate,d0
skip3
			cmp.w	temp,d0					;             && (A[j] > temp) ) {
			bls.s	endWhile2
			subq	#2,j_indx				;	      --j;
			bra.s	while2					;      }
endWhile2:
			cmp.w	i_indx,j_indx			;      if (i != j) {
			beq.s	endIf1
			move.w	0(array,j_indx.l),0(array,i_indx.l)  ;  A[i] = A[j];
			addq.w	#2,i_indx				;           ++i;
			
while3:		cmp.w	i_indx,j_indx			;           while ( (i != j) 
			beq.s	endWhile3				;  

			move.w	0(array,i_indx.l),d0
			tst.b	rotate
			beq.s	skip4
			rol.w	rotate,d0
skip4
			cmp.w	temp,d0            		;              && (A[i] < temp) {
			bhs.s	endWhile3
			addq.w	#2,i_indx				;				  ++i;
			bra.s	while3					;            }
endWhile3:
			cmp.w	i_indx,j_indx			;            if (i != j) {
			beq.s	endIf2
			move.w	0(array,i_indx.l),0(array,j_indx.l)  ;       A[j] = A[i];
			subq.w	#2,j_indx				;                --j;
endIf2:										;             }
endIf1:										;        }

			cmp.w	i_indx,j_indx		
			bne.s	doWhile					;  } while (i != j);

			move.w	temp_indx,0(array,i_indx.l) ;  A[i] = temp;
			
			cmp.w	j_indx,first			; if (first < j)
			bge.s	skipSort1				
			move.w	last,-(sp)
			move.w	j_indx,last
			subq.w	#2,last
			jsr		myqsort					;     quickSort(A[], first, j-1);
			move.w	(sp)+,last
skipSort1

			cmp.w	last,i_indx				; if (i < last)
			bge.s	skipSort2
			move.w	i_indx,first
			addq.w	#2,first
			jsr		myqsort					;     quickSort(A[], i+1, last);
skipSort2

exit:			
			movem.w	(sp)+,d2-d7
			rts
			ENDPROC
			
*****************************************************************************
* NAME: bldPlut
* DESC: Fill in the PLUT with color indices
* CDEF:	void bldPlut(byte *plutAddr);
* REGS: 
*****************************************************************************
bldPlut	PROC	EXPORT
			
frame	RECORD	{A6Link},DECR
plutAddr	ds.l	1
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0-a4,-(sp)
			
			move.l	frame.plutAddr(a6),a0
			lea		locals.crOffset,a2		; a2 = base of CR offsets
			move.l	locals.rectAddr,a1		; a1 = base of CR array
			move.w	#0,d1					; d1 = current CR entry
			move.w	locals.numCRused,d2		
			subq.w	#1,d2
			asl.w	#1,d2					; d2 = last CR entry
			move.w	#1,d7					; d7 = palette index
			
loop
			cmp.w	d2,d1
			bgt.s	exit
			clr.l	d0
			move.w	0(a2,d1.w),d5
			lea		0(a1,d5.w),a3
			move.b	d7,colorRect.palIndex(a3)
			call	bldOnePlut
			addq.w	#2,d1
			addq.w	#1,d7
			bra.s	loop

exit			
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC
*****************************************************************************
* NAME: bldOnePlut
* DESC: Initialize the PLUT with index entries for one color
* REGS: 
*****************************************************************************
bldOnePlut	PROC	EXPORT
			
frame	RECORD	{A6Link},DECR
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0-a4,-(sp)
			
			WITH	colorRect
				
			clr.l	d0
			clr.l	d1
			clr.l	d2
			clr.l	d3
			clr.l	d4
			clr.l	d5

			move.b	redMin(a3),d0
			move.b	redMax(a3),d1
			move.b	greenMax(a3),d3
			move.b	blueMax(a3),d5
			
redLoop
			move.b	greenMin(a3),d2
greenLoop
			move.b	blueMin(a3),d4
			
			move.b	d0,d6
			lsl.w	#5,d6
			or.b	d2,d6
			lsl.w	#5,d6
			or.b	d4,d6
			
blueLoop
			move.b	d7,0(a0,d6.l)
			addq.b	#1,d6
			addq.b	#1,d4
			cmp.b	d4,d5
			bge.s	blueLoop
			
			addq.b	#1,d2
			cmp.b	d2,d3
			bge.s	greenLoop
			
			addq.b	#1,d0
			cmp.b	d0,d1
			bge.s	redLoop

exit			
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC

*****************************************************************************
* NAME: calcCRcolor
* DESC: Determine the appropriate color for each colorRect entry
* CDEF:	void calcCRcolor();
* REGS: 
*****************************************************************************
calcCRcolor	PROC	EXPORT
			
frame	RECORD	{A6Link},DECR
retAddr		ds.l	1
A6Link		ds.l	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0-a4,-(sp)
			
			lea		locals.crOffset,a0		; a0 = base of CR offsets
			move.l	locals.rectAddr,a1		; a1 = base of CR array
			move.l	locals.listAddr,a2		; a2 = base of colorList array
			move.l	locals.histAddr,a4		; a4 = base of histogram array
			
			move.w	#0,d1					; d1 = current CR entry
			move.w	locals.numCRused,d2
			subq.w	#1,d2
			asl.w	#1,d2					; d2 = last CR entry
	
loop
			cmp.w	d1,d2
			blt.s	ok
			move.w	0(a0,d1.w),d5
			lea		0(a1,d5.w),a3
			call	calcOneColor
			addq.w	#2,d1
			bra.s	loop

ok			move.l	#OK,d0

exit			
			movem.l	(sp)+,d1-d7/a0-a4
			unlk	a6
			rts
			ENDPROC
*****************************************************************************
* NAME: calcOneColor
* DESC: 
* CDEF:	
* RETN:	
* REGS: On entry: 
*			a2	colorList array
*			a3	current colorRect entry
*			a4	histogram array
*****************************************************************************
calcOneColor	PROC	EXPORT
			
frame	RECORD	{A6Link},DECR
retAddr		ds.l	1
A6Link		ds.l	1
crEnd		ds.l	1
color		ds.w	1
size		equ		*
			ENDR
			
			link	a6,#frame.size
			movem.l	d1-d7/a0-a1,-(sp)
			
			WITH	colorRect
			move.l	first(a3),d3
			move.l	last(a3),d0
			lsl.l	#1,d3
			lsl.l	#1,d0
			move.l	d0,frame.crEnd(a6)
			
			clr.l	d4						; d4 = sum of (red * histogram)
			clr.l	d5						; d5 = sum of (green * histogram)			
		
			clr.l	d6						; d6 = sum of (blue * histogram)
			
			
loop
			clr.l	d1
			clr.l	d2
			clr.l	d7
			move.w	0(a2,d3.l),d2			; get color (15-bit RGB) value
			lsl.l	#1,d2					;   histogram word offset
			move.w	0(a4,d2.l),d1			;   d1 = histogram value
			lsr.l	#1,d2
			move.w	d2,frame.color(a6)
			
			clr.l	d0
			clr.l	d7
			move.b	d2,d0					; get blue value
			andi.b	#$1F,d0					;   (low 5 bits)
			mulu	d1,d0					;   multiply by histogram
			add.l	d0,d6					;     and add to summary
			
			clr.l	d0
			clr.l	d7
			lsr.w	#5,d2
			move.b	d2,d0					; get green value
			andi.b	#$1F,d0					;   (low 5 bits)
			mulu	d1,d0					;   multiply by histogram
			add.l	d0,d5					;     and add to summary
			
			clr.l	d0
			clr.l	d7
			lsr.w	#5,d2
			move.b	d2,d0					; get red value
			andi.b	#$1F,d0					;   (low 5 bits)
			mulu	d1,d0					;   multiply by histogram
			add.l	d0,d4					;     and add to summary
			
			move.w	d3,d2
			lsr.w	#1,d2
				
			addq.l	#2,d3
			cmp.l	frame.crEnd(a6),d3
			ble.s	loop	
			
			move.l	histSum(a3),d2
			cmp.l	#$0000FFFF,d2
			bls.s	withinRange
			move.l	#$0000FFFF,d2
withinRange			
			move.l	d2,d1			; To get actual color values
			lsr.l	#1,d1			; we'll divide each color sum
			add.l	d1,d4			; by the histogram.  But since it's
			add.l	d1,d5			; an integer divide, we need to add
			add.l	d1,d6			; half the histogram first.
			
			tst.w	d4
			beq.s	skipRed
			divu	d2,d4			; Now we divide red
skipRed
			tst.w	d5
			beq.s	skipGreen
			divu	d2,d5			; Now we divide green
skipGreen
			tst.w	d6
			beq.s	skipBlue
			divu	d2,d6			; Now we divide blue
skipBlue
			
			move.b	d4,red(a3)
			move.b	d5,green(a3)
			move.b	d6,blue(a3)
			clr.w	d0
exit			
			movem.l	(sp)+,d1-d7/a0-a1
			unlk	a6
			rts
			ENDPROC
			
			END
